// Map.cc

#include "Map.h"
#include "SFMT.h"
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <string>

int Map::map_width, Map::map_height;
uint64_t Map::hashkeys_moving;
uint64_t* Map::hashkeys_walls;
uint64_t* Map::hashkeys_bikes;

uint64_t wall_allocated;
uint64_t** wall_blocks;
void* wall_arena;
void* wall_arena_end;

void* align16(void* ptr)
{
	return (void*)((size_t(ptr) + 0xF) & ~0xF);
}

int bitindex(uint64_t bit)
{
    int index = ((bit & 0xAAAAAAAAAAAAAAAAull) != 0);
    index |= ((bit & 0xCCCCCCCCCCCCCCCCull) != 0) << 1;
    index |= ((bit & 0xF0F0F0F0F0F0F0F0ull) != 0) << 2;
    index |= ((bit & 0xFF00FF00FF00FF00ull) != 0) << 3;
    index |= ((bit & 0xFFFF0000FFFF0000ull) != 0) << 4;
    index |= ((bit & 0xFFFFFFFF00000000ull) != 0) << 5;
    return index;
}

void Map::preallocate()
{
    wall_allocated = 0;
    wall_blocks = new uint64_t*[64];
    wall_arena = malloc(64 * ((sizeof(uint64_t) * map_width) + 15));
    wall_arena_end = (char*)wall_arena
        + (64 * ((sizeof(uint64_t) * map_width) + 15));
    //fprintf(stderr, "arena start %X end %X ", wall_arena, wall_arena_end);
    for (int block = 0; block < 64; block++)
    {
        char* block_start = (char*)wall_arena + (block
                * ((sizeof(uint64_t) * map_width) + 15));
        uint64_t* wall = (uint64_t*)align16(block_start);
        wall_blocks[block] = wall;
    }
    /*
    fprintf(stderr, "last wall %X\n",
            &wall_blocks[63][map_width-1][map_height-1]);
    */
}

uint64_t* Map::newWalls()
{
    if (~wall_allocated)
    {
        uint64_t block_bit = ~wall_allocated & -(~wall_allocated);
        wall_allocated ^= block_bit;
        int block = bitindex(block_bit);
        memset(wall_blocks[block], 0, sizeof(uint64_t) * map_width);
        return wall_blocks[block];
    } else {
        uint64_t* walls = new uint64_t[map_width];
        memset(walls, 0, sizeof(uint64_t) * map_width);
        return walls;
    }
}

uint64_t* Map::newWalls(uint64_t const * const initial)
{
    if (~wall_allocated)
    {
        uint64_t block_bit = ~wall_allocated & -(~wall_allocated);
        wall_allocated ^= block_bit;
        int block = bitindex(block_bit);
        memcpy(wall_blocks[block], initial, sizeof(uint64_t) * map_width);
        return wall_blocks[block];
    } else {
        uint64_t* walls = new uint64_t[map_width];
        memcpy(walls, initial, map_width);
        return walls;
    }
}

void Map::freeWalls(uint64_t* walls)
{
    if (walls >= wall_arena && walls < wall_arena_end)
    {
        for (int i=0; i < 64; i++)
        {
            if (wall_blocks[i] == walls)
            {
                wall_allocated &= ~(1 << i);
                break;
            }
            if (i == 63)
            {
                fprintf(stderr, "wall block not found");
                exit(1);
            }
        }
    } else {
        delete[] walls;
    }
}

Map::Map() {
  ReadFromFile(stdin);
}

Map::Map(const Map& other)
{
    moving = other.moving;
    bikes[0][0] = other.bikes[0][0];
    bikes[0][1] = other.bikes[0][1];
    bikes[1][0] = other.bikes[1][0];
    bikes[1][1] = other.bikes[1][1];
    hash = other.hash;
    walls = newWalls(other.walls);
}

Map::~Map()
{
    freeWalls(walls);
}

bool Map::operator==(const Map& other) const
{
    if (this == &other)
        return true;
    if (hash != other.hash
            || moving != other.moving)
        return false;
    if (bikes[0][0] != other.bikes[0][0]
            || bikes[0][1] != other.bikes[0][1]
            || bikes[1][0] != other.bikes[1][0]
            || bikes[1][1] != other.bikes[1][1])
        return false;
    if (memcmp(walls, other.walls, sizeof(uint64_t) * map_width))
    {
        return false;
    }
    return true;
}

bool Map::isWall(const int* const cord) const {
  if (cord[0] < 0 || cord[1] < 0
          || cord[0] >= map_width || cord[1] >= map_height) {
      return true;
  } else {
      return (walls[cord[0]] & (UINT64_C(1) << cord[1])) != 0;
  }
}

bool Map::isWall(int x, int y) const
{
    if (x < 0 || y < 0 || x >= map_width || y >= map_height)
    {
        return true;
    }
    return (walls[x] & (UINT64_C(1) << y)) != 0;
}

void Map::toString(std::string& board)
{
    for (int j=0; j < map_height; j++)
    {
        for (int i=0; i < map_width; i++)
        {
            bool at_bike1 = bikes[0][0] == i && bikes[0][1] == j;
            bool at_bike2 = bikes[1][0] == i && bikes[1][1] == j;

            if (at_bike1 && !at_bike2)
            {
                if (isWall(i, j))
                    board.append("!");
                else
                    board.append("1");
            }
            else if (at_bike2 && !at_bike1)
            {
                if (isWall(i, j))
                    board.append("@");
                else
                    board.append("2");
            }
            else if (at_bike1 && at_bike2)
            {
                board.append("B");
            }
            else if (isWall(i, j))
            {
                board.append("#");
            } else {
                board.append(" ");
            }
        }
        board.append("\n");
    }
}

void Map::makeMove(const std::string& move) {
  if (move.length() == 0) {
    fprintf(stderr, "ERROR: zero-length string passed to MakeMove(string)\n");
    exit(1);
  } else {
    int c = (int)move[0];
    switch (c) {
    case 'n':
    case 'N':
      makeMove(0);
      break;
    case 'e':
    case 'E':
      makeMove(1);
      break;
    case 's':
    case 'S':
      makeMove(2);
      break;
    case 'w':
    case 'W':
      makeMove(3);
      break;
    default:
      fprintf(stderr, "Invalid string passed to MakeMove(string): %s\n"
            "Move string must start with N, E, S, or W!", move.c_str());
      exit(1);
      break;
    }
  }
}

void Map::makeMove(int move) {
  fprintf(stdout, "%d\n", move+1);
  fflush(stdout);
}

void Map::doPlayoutStep(int direction)
{
    int* cur_bike = bikes[moving];
    walls[cur_bike[0]] |= UINT64_C(1) << cur_bike[1];
    switch (direction)
    {
        case 0:
            cur_bike[1] -= 1;
            break;
        case 1:
            cur_bike[0] += 1;
            break;
        case 2:
            cur_bike[1] += 1;
            break;
        case 3:
            cur_bike[0] -= 1;
            break;
        default:
            fprintf(stderr, "Bad direction given to doPlayoutStep\n");
            break;
    }
    moving ^= 1;
}

void Map::doStep(int direction)
{
    int startx = bikes[moving][0];
    int starty = bikes[moving][1];
    hash ^= hashkeys_bikes[(moving * map_width * map_height)
        + (startx * map_height) + starty];
    walls[startx] |= UINT64_C(1) << starty;
    hash ^= hashkeys_walls[(startx * map_height) + starty];
    switch(direction)
    {
        case 0:
            bikes[moving][1] -= 1;
            break;
        case 1:
            bikes[moving][0] += 1;
            break;
        case 2:
            bikes[moving][1] += 1;
            break;
        case 3:
            bikes[moving][0] -= 1;
            break;
    }
    int finalx = bikes[moving][0];
    int finaly = bikes[moving][1];
    hash ^= hashkeys_bikes[(moving * map_width * map_height)
        + (finalx * map_height) + finaly];
    moving = moving ^ 1;
    hash ^= hashkeys_moving;
}

void Map::undoStep(int direction)
{
    moving = moving ^ 1;
    hash ^= hashkeys_moving;
    int startx = bikes[moving][0];
    int starty = bikes[moving][1];
    hash ^= hashkeys_bikes[(moving * map_width * map_height)
        + (startx * map_height) + starty];
    switch(direction)
    {
        case 0:
            bikes[moving][1] += 1;
            break;
        case 1:
            bikes[moving][0] -= 1;
            break;
        case 2:
            bikes[moving][1] -= 1;
            break;
        case 3:
            bikes[moving][0] += 1;
            break;
    }
    int finalx = bikes[moving][0];
    int finaly = bikes[moving][1];
    walls[finalx] &= ~(1 << finaly);
    hash ^= hashkeys_walls[(finalx * map_height) + finaly];
    hash ^= hashkeys_bikes[(moving * map_width * map_height)
        + (finalx * map_height) + finaly];
}

bool Map::isEnd() const
{
    if (moving)
        return false;
    if (bikes[0][0] == bikes[1][0] && bikes[0][1] == bikes[1][1])
        return true;
    if ((walls[bikes[0][0]] & (UINT64_C(1) << bikes[0][1]))
            || (walls[bikes[1][0]] & (UINT64_C(1) << bikes[1][1])))
        return true;
    return false;
}

double Map::score() const
{
    if (moving)
        return 1/2.0;
    bool myCrash = (walls[bikes[0][0]] & (UINT64_C(1) << bikes[0][1])) != 0;
    bool oppCrash = (walls[bikes[1][0]] & (UINT64_C(1) << bikes[1][1])) != 0;
    if (myCrash && !oppCrash)
        return 0;
    if (oppCrash && !myCrash)
        return 1;
    return 1/2.0;
}

void Map::ReadFromFile(FILE *file_handle) {
  int x, y, c;
  int width, height;
  int num_items = fscanf(file_handle, "%d %d\n", &width, &height);
  if (feof(file_handle) || num_items < 2) {
    exit(0); // End of stream means end of game. Just exit.
  }
  if (map_width != width || map_height != height)
  {
      if (map_width || map_height)
      {
          fprintf(stderr, "Map size changed from previous size\n");
          exit(1);
      }
      map_width = width;
      map_height = height;
      generateKeys();
      preallocate();
  }
  walls = newWalls();
  moving = 0;
  x = 0;
  y = 0;
  while (y < map_height && (c = fgetc(file_handle)) != EOF) {
    switch (c) {
    case '\r':
      break;
    case '\n':
      if (x != map_width) {
        fprintf(stderr, "x != width in Board_ReadFromStream\n");
        return;
      }
      ++y;
      x = 0;
      break;
    case '#':
      if (x >= map_width) {
        fprintf(stderr, "x >= width in Board_ReadFromStream\n");
        return;
      }
      walls[x] |= UINT64_C(1) << y;
      ++x;
      break;
    case ' ':
      if (x >= map_width) {
        fprintf(stderr, "x >= width in Board_ReadFromStream\n");
        return;
      }
      walls[x] &= ~(UINT64_C(1) << y);
      ++x;
      break;
    case '1':
      if (x >= map_width) {
        fprintf(stderr, "x >= width in Board_ReadFromStream\n");
        return;
      }
      walls[x] &= ~(UINT64_C(1) << y);
      bikes[0][0] = x;
      bikes[0][1] = y;
      ++x;
      break;
    case '2':
      if (x >= map_width) {
        fprintf(stderr, "x >= width in Board_ReadFromStream\n");
        return;
      }
      walls[x] &= ~(UINT64_C(1) << y);
      bikes[1][0] = x;
      bikes[1][1] = y;
      ++x;
      break;
    default:
      fprintf(stderr, "unexpected character %d in Board_ReadFromStream", c);
      return;
    }
  }
  generateHash();
}

void Map::generateHash()
{
    hash = 0;
    for (int i=0; i < map_width; i++)
    {
        for (int j=0; j < map_height; j++)
        {
			if (walls[i] & (UINT64_C(1) << j))
                hash ^= hashkeys_walls[(i * map_height) + j];
        }
    }
    for (int i=0; i < 2; i++)
    {
        int x = bikes[i][0];
        int y = bikes[i][1];
        hash ^= hashkeys_bikes[(i * map_width * map_height)
            + (x*map_height) + y];
    }
    if (moving)
        hash ^= hashkeys_moving;
}

void Map::generateKeys()
{
    hashkeys_moving = gen_rand64();
    hashkeys_walls = new uint64_t[map_width * map_height];
    for (int i=0; i < (map_width * map_height); i++)
    {
        hashkeys_walls[i] = gen_rand64();
    }
    hashkeys_bikes = new uint64_t[2 * map_width * map_height];
    for (int i=0; i < (2 * map_width * map_height); i++)
    {
        hashkeys_bikes[i] = gen_rand64();
    }
}
