// Map.h
//
// Handles the Tron map. Also handles communicating with the Tron game engine.
// You don't need to change anything in this file.

#include <string>

#if defined(_MSC_VER)
  typedef unsigned __int64 uint64_t;
  #define UINT64_C(c) c ## ULL
#else
  #include <inttypes.h>
#endif

#pragma once

class Map {
 public:
  // Constructs a Map by reading an ASCII representation from the console
  // (stdin).
  Map();

  // Copy contructor
  Map(const Map& other);

  ~Map();

  // Returns whether or not the given cell is a wall or not. TRUE means it's
  // a wall, FALSE means it's not a wall, and is passable. Any spaces that are
  // not on the board are deemed to be walls.
  bool isWall(const int* const cord) const;
  bool isWall(int x, int y) const;

  void toString(std::string& board);

  // Sends your move to the contest engine. Only the first character of
  // the string is used. It is case insensitive. The four possible moves are
  //   * "N" -- North. Negative Y direction.
  //   * "E" -- East. Positive X direction.
  //   * "S" -- South. Positive X direction.
  //   * "W" -- West. Negative X direction.
  // Other strings can be valid moves, too. For example: "North", "east",
  // "s", and "WwWwWest!" are all valid moves, because they start with one
  // of the four allowed characters.
  static void makeMove(const std::string& move);

  // Sends your move to the contest engine. The four possible moves are
  //   * 1 -- North. Negative Y direction.
  //   * 2 -- East. Positive X direction.
  //   * 3 -- South. Positive X direction.
  //   * 4 -- West. Negative X direction.
  static void makeMove(int move);

  // Make a step without updating the hash
  void doPlayoutStep(int direction);

  // Make a step
  void doStep(int direction);

  // Undo a step
  void undoStep(int direction);

  bool isEnd() const;
  double score() const;

  bool operator==(const Map& other) const;
  bool operator!=(const Map& other) const { return !(*this == other); }

  // Map dimensions.
  static int map_width, map_height;

  // The locations of both players.
  int bikes[2][2];

  // side moving
  int moving;

  // position hash
  uint64_t hash;

 private:
  // Load a board from an open file handle. To read from the console, pass
  // stdin, which is actually a (FILE*).
  //   file_handle -- an open file handle from which to read.
  //
  // If there is a problem, the function returns NULL. Otherwise, a valid
  // Board structure is returned.
  //
  // The file should be an ascii file. The first line contains the width and
  // height of the board, separated by a space. subsequent lines contain visual
  // representations of the rows of the board, using '#' and space characters.
  // The starting positions of the two players are indicated by '1' and '2'
  // characters. There must be exactly one '1' character and one '2' character
  // on the board. For example:
  // 6 4
  // ######
  // #1# 2#
  // #   ##
  // ######
  void ReadFromFile(FILE *file_handle);

  void generateHash();

  static void generateKeys();
  static void preallocate();

  static uint64_t* newWalls();
  static uint64_t* newWalls(uint64_t const * const walls);
  static void freeWalls(uint64_t* walls);

  // the walls on the board in bit board form for each column
  uint64_t* walls;

  static uint64_t hashkeys_moving;
  static uint64_t* hashkeys_walls;
  static uint64_t* hashkeys_bikes;
};
